#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import os
import sys
import uuid
import plistlib
import platform

from ak_vendor.usbmux import default_mux as mux

from pymobiledevice.ca import ca_do_everything
from pymobiledevice.plist_service import PlistService
from pymobiledevice.util import readHomeFile, writeHomeFile


class NotTrustedError(Exception):
    pass


class PairingError(Exception):
    pass


class NotPairedError(Exception):
    pass


class CannotStopSessionError(Exception):
    pass


class StartServiceError(Exception):

    def __init__(self, message):
        print("[ERROR] %s" % message)


class FatalPairingError(Exception):
    pass


# we store pairing records and ssl keys in ~/.pymobiledevice
HOMEFOLDER = ".pymobiledevice"
MAXTRIES = 20


def list_devices():
    mux.process(1)
    return [d.serial for d in mux.devices]


class LockdownClient(object):

    def __init__(self, udid=None):
        self.paired = False
        self.SessionID = None
        self.c = PlistService(62078, udid)
        self.hostID = self.generate_hostID()
        self.SystemBUID = self.generate_hostID()
        self.paired = False
        self.label = "pyMobileDevice"

        assert self.queryType() == "com.apple.mobile.lockdown"

        self.udid = udid
        self.allValues = self.getValue()
        self.UniqueChipID = self.allValues.get("UniqueChipID")
        self.DevicePublicKey = self.getValue("", "DevicePublicKey")
        self.identifier = self.udid
        if not self.identifier:
            if self.UniqueChipID:
                self.identifier = "%x" % self.UniqueChipID
            else:
                print("Could not get UDID or ECID, failing")
                raise

        if not self.validate_pairing():
            self.pair()
            if not self.validate_pairing():
                raise FatalPairingError
            self.paired = True
        return

    def queryType(self):
        self.c.sendPlist({"Request": "QueryType"})
        res = self.c.recvPlist()
        return res.get("Type")

    def generate_hostID(self):
        hostname = platform.node()
        hostid = uuid.uuid3(uuid.NAMESPACE_DNS, hostname)
        return str(hostid).upper()

    def enter_recovery(self):
        self.c.sendPlist({"Request": "EnterRecovery"})
        print(self.c.recvPlist())

    def stop_session(self):
        if self.SessionID and self.c:
            self.c.sendPlist({
                "Label": self.label,
                "Request": "StopSession",
                "SessionID": self.SessionID})
            self.SessionID = None
            res = self.c.recvPlist()
            if not res or res.get("Result") != "Success":
                raise CannotStopSessionError
            return res

    def validate_pairing(self):
        pair_record = None
        certPem = None
        privateKeyPem = None

        if sys.platform == "win32":
            folder = os.environ["ALLUSERSPROFILE"] + "/Apple/Lockdown/"
        elif sys.platform == "darwin":
            folder = "/var/db/lockdown/"
        elif len(sys.platform) >= 5:
            if sys.platform[0:5] == "linux":
                folder = "/var/lib/lockdown/"
        try:
            pair_record = plistlib.readPlist(
                folder + "%s.plist" % self.identifier)
        except:
            pair_record = None
        if pair_record:
            print("Using iTunes pair record: %s.plist" % self.identifier)
            certPem = pair_record["HostCertificate"].data
            privateKeyPem = pair_record["HostPrivateKey"].data

        else:
            print("No iTunes pairing record found for device %s" %
                  self.identifier)
            print("Looking for pymobiledevice pairing record")
            record = readHomeFile(HOMEFOLDER, "%s.plist" % self.identifier)
            if record:
                pair_record = plistlib.loads(record)
                print("Found pymobiledevice pairing record for device %s" %
                      self.udid)
                certPem = pair_record["HostCertificate"].data
                privateKeyPem = pair_record["HostPrivateKey"].data
            else:
                print("No  pymobiledevice pairing record found for device %s" %
                      self.identifier)
                return False

        ValidatePair = {"Label": self.label,
                        "Request": "ValidatePair", "PairRecord": pair_record}
        self.c = PlistService(62078, self.udid)
        self.c.sendPlist(ValidatePair)
        r = self.c.recvPlist()
        if not r or "Error" in r:
            pair_record = None
            print("ValidatePair fail", ValidatePair)
            return False

        self.hostID = pair_record.get("HostID", self.hostID)
        self.SystemBUID = pair_record.get("SystemBUID", self.SystemBUID)
        d = {"Label": self.label, "Request": "StartSession",
             "HostID": self.hostID, 'SystemBUID': self.SystemBUID}
        self.c.sendPlist(d)
        startsession = self.c.recvPlist()
        self.SessionID = startsession.get("SessionID")
        if startsession.get("EnableSessionSSL"):
            sslfile = self.identifier + "_ssl.txt"
            # print(HOMEFOLDER, sslfile, certPem, "\n", privateKeyPem)
            sslfile = writeHomeFile(
                HOMEFOLDER, sslfile, certPem + b"\n" + privateKeyPem)
            self.c.ssl_start(sslfile, sslfile)

        self.paired = True
        return True

    def pair(self):
        self.DevicePublicKey = self.getValue("", "DevicePublicKey")
        if self.DevicePublicKey == '':
            print("Unable to retreive DevicePublicKey")
            return False

        print("Creating host key & certificate")
        certPem, privateKeyPem, DeviceCertificate = ca_do_everything(
            self.DevicePublicKey)

        pair_record = {"DevicePublicKey": plistlib.Data(self.DevicePublicKey),
                       "DeviceCertificate": plistlib.Data(DeviceCertificate),
                       "HostCertificate": plistlib.Data(certPem),
                       "HostID": self.hostID,
                       "RootCertificate": plistlib.Data(certPem),
                       "SystemBUID": "30142955-444094379208051516"}

        pair = {"Label": self.label, "Request": "Pair",
                "PairRecord": pair_record}
        self.c = PlistService(62078, self.udid)
        self.c.sendPlist(pair)
        pair = self.c.recvPlist()

        if pair and pair.get("Result") == "Success" or "EscrowBag" in pair:
            pair_record["HostPrivateKey"] = plistlib.Data(privateKeyPem)
            pair_record["EscrowBag"] = pair.get("EscrowBag")
            writeHomeFile(HOMEFOLDER, "%s.plist" %
                          self.identifier, plistlib.dumps(pair_record))
            self.paired = True
            return True

        elif pair and pair.get("Error") == "PasswordProtected":
            self.c.close()
            raise NotTrustedError

        else:
            print(pair.get("Error"))
            self.c.close()
            raise PairingError

    def getValue(self, domain=None, key=None):

        if(
                isinstance(key, str) and
                hasattr(self, 'record')and
                hasattr(self.record, key)):
            return self.record[key]

        req = {"Request": "GetValue", "Label": self.label}

        if domain:
            req["Domain"] = domain
        if key:
            req["Key"] = key

        self.c.sendPlist(req)
        res = self.c.recvPlist()
        if res:
            r = res.get("Value")
            if hasattr(r, "data"):
                return r.data
            return r

    def setValue(self, domain=None, key=None):

        req = {"Request": "SetValue", "Label": self.label}

        if domain:
            req["Domain"] = domain
        if key:
            req["Key"] = key

        self.c.sendPlist(req)
        res = self.c.recvPlist()
        print(res)
        return res

    def startService(self, name):
        if not self.paired:
            print("NotPaired")
            raise NotPairedError

        self.c.sendPlist(
            {"Label": self.label, "Request": "StartService", "Service": name})
        StartService = self.c.recvPlist()
        if not StartService or StartService.get("Error"):
            raise StartServiceError(StartService.get("Error"))
        return PlistService(StartService.get("Port"), self.udid)

    def startServiceWithEscrowBag(self, name, escrowBag=None):
        if not self.paired:
            print("NotPaired")
            raise NotPairedError

        if (not escrowBag):
            escrowBag = self.record['EscrowBag']

        self.c.sendPlist({
            "Label": self.label, "Request": "StartService", "Service": name,
            'EscrowBag': plistlib.Data(escrowBag)})
        StartService = self.c.recvPlist()
        if not StartService or StartService.get("Error"):
            raise StartServiceError(StartService.get("Error"))
        return PlistService(StartService.get("Port"), self.udid)


if __name__ == "__main__":
    l = LockdownClient()
    if l:
        n = writeHomeFile(HOMEFOLDER, "%s_infos.plist" %
                          l.udid, plistlib.dumps(l.allValues))
        print("Wrote infos to %s" % n)
    else:
        print("Unable to connect to device")



import os
import sys
from util import sizeof_fmt, hexdump
from progressbar import ProgressBar
from crypto.aes import AESdecryptCBC, AESencryptCBC


class FileBlockDevice(object):

    def __init__(self, filename, offset=0, write=False):
        flag = os.O_RDONLY if not write else os.O_RDWR
        if sys.platform == 'win32':
            flag = flag | os.O_BINARY
        self.filename = filename
        self.fd = os.open(filename, flag)
        self.offset = offset
        self.writeFlag = write
        self.size = os.path.getsize(filename)
        self.setBlockSize(8192)

    def setBlockSize(self, bs):
        self.blockSize = bs
        self.nBlocks = self.size / bs

    def readBlock(self, blockNum):
        os.lseek(self.fd, self.offset + self.blockSize * blockNum, os.SEEK_SET)
        return os.read(self.fd, self.blockSize)

    def write(self, offset, data):
        if self.writeFlag:  # fail silently for testing
            os.lseek(self.fd, self.offset + offset, os.SEEK_SET)
            return os.write(self.fd, data)

    def writeBlock(self, lba, block):
        return self.write(lba * self.blockSize, block)


class FTLBlockDevice(object):

    def __init__(self, nand, first_lba, last_lba, defaultKey=None):
        self.nand = nand
        self.pageSize = nand.pageSize
        self.blockSize = 0  # not used
        self.key = defaultKey
        self.lbaoffset = first_lba
        self.last_lba = last_lba
        self.setBlockSize(self.pageSize)

    def setBlockSize(self, bs):
        self.blockSize = bs
        self.lbasPerPage = self.pageSize / bs
        self.lbaToLpnFactor = bs / (self.pageSize + 0.0)
        self.pagesPerLBA = bs / self.pageSize
        if bs > self.pageSize:
            # raise Exception("FTLBlockDevice lba-size > pageSize not handled")
            pass

    def readBlock(self, blockNum):
        # if (self.lbaoffset + blockNum / self.lbasPerPage) > self.last_lba:
        #    print "readBlock past last lba", blockNum
        #    print "readBlock past last lba", blockNum
        #    return "\x00" * self.blockSize
        lpn = int(self.lbaoffset + blockNum * self.lbaToLpnFactor)
        d = self.nand.readLPN(lpn, self.key)
        for i in range(1, self.pagesPerLBA):
            d += self.nand.readLPN(lpn + i, self.key)
        if self.lbasPerPage:
            zz = blockNum % self.lbasPerPage
            return d[zz * self.blockSize:(zz + 1) * self.blockSize]
        return d

    def write(self, offset, data):
        raise Exception("FTLBlockDevice write method not implemented")

    def writeBlock(self, lba, block):
        raise Exception("FTLBlockDevice writeBlock method not implemented")

    def dumpToFile(self, outputfilename):
        hs = sizeof_fmt((self.last_lba - self.lbaoffset) * self.pageSize)
        print("Dumping partition to %s (%s)" % (outputfilename, hs))
        flags = os.O_CREAT | os.O_RDWR
        if sys.platform == "win32":
            flags |= os.O_BINARY
        fd = os.open(outputfilename, flags)

        pbar = ProgressBar(self.last_lba - self.lbaoffset - 1)
        pbar.start()
        for i in range(self.lbaoffset, self.last_lba):
            pbar.update(i - self.lbaoffset)
            d = self.nand.readLPN(i, self.key)
            if i == self.lbaoffset and d[0x400:0x402] != "HX":
                print("FAIL? Not HFS partition or wrong key")
            os.write(fd, d)
        pbar.finish()
        os.close(fd)


class IMG3BlockDevice(object):

    def __init__(self, filename, key, iv, write=False):
        flag = os.O_RDONLY if not write else os.O_RDWR
        if sys.platform == 'win32':
            flag = flag | os.O_BINARY
        self.filename = filename
        self.fd = os.open(filename, flag)
        self.writeFlag = write
        d = os.read(self.fd, 8192)
        if d[:4] != "3gmI":
            raise Exception("IMG3BlockDevice bad magic %s" % d[:4])
        if d[0x34:0x38] != "ATAD":
            raise Exception("Fu")
        self.encrypted = True
        self.key = key
        self.iv0 = iv
        self.offset = 0x40
        self.size = os.path.getsize(filename)
        self.setBlockSize(8192)

    def setBlockSize(self, bs):
        self.blockSize = bs
        self.nBlocks = self.size / bs
        self.ivs = {0: self.iv0}

    def getIVforBlock(self, blockNum):
        # read last 16 bytes of previous block to get IV
        if blockNum not in self.ivs:
            os.lseek(self.fd, self.offset + self.blockSize *
                     blockNum - 16, os.SEEK_SET)
            self.ivs[blockNum] = os.read(self.fd, 16)
        return self.ivs[blockNum]

    def readBlock(self, blockNum):
        os.lseek(self.fd, self.offset + self.blockSize * blockNum, os.SEEK_SET)
        data = os.read(self.fd, self.blockSize)
        if self.encrypted:
            data = AESdecryptCBC(data, self.key, self.getIVforBlock(blockNum))
        return data

    def _write(self, offset, data):
        if self.writeFlag:  # fail silently for testing
            os.lseek(self.fd, self.offset + offset, os.SEEK_SET)
            return os.write(self.fd, data)

    def writeBlock(self, lba, data):
        if self.encrypted:
            data = AESencryptCBC(data, self.key, self.getIVforBlock(lba))
        return self._write(lba * self.blockSize, data)



from bsdiff import Patch
import os
import sys
import bz2
from array import array
import struct
import io as StringIO


def patch(old_file_name, new_file_name, patch_file_name):

    patch_file = open(patch_file_name, "rb")
    patch_file.read(8)  # magic number

    compressed_control_len = offtin(patch_file.read(8))
    compressed_diff_len = offtin(patch_file.read(8))
    new_file_len = offtin(patch_file.read(8))

    compressed_control_block = patch_file.read(compressed_control_len)
    compressed_diff_block = patch_file.read(compressed_diff_len)
    compressed_extra_block = patch_file.read()

    control_stream = StringIO.StringIO(
        bz2.decompress(compressed_control_block))
    diff_string = bz2.decompress(compressed_diff_block)
    extra_string = bz2.decompress(compressed_extra_block)

    control_tuples_list = []
    while True:
        r = control_stream.read(8)
        if not r:
            break
        x = offtin(r)
        y = offtin(control_stream.read(8))
        z = offtin(control_stream.read(8))
        control_tuples_list.append((x, y, z))

    old_data = open(old_file_name, "rb").read()
    new_data = Patch(old_data, new_file_len,
                     control_tuples_list, diff_string, extra_string)

    new_file = open(new_file_name, "wb")
    new_file.write(new_data)


def offtin(buf):
    s = ord(buf[7])

    y = (s & 0x7F) << 56
    y += ord(buf[6]) << 48
    y += ord(buf[5]) << 40
    y += ord(buf[4]) << 32
    y += ord(buf[3]) << 24
    y += ord(buf[2]) << 16
    y += ord(buf[1]) << 8
    y += ord(buf[0])

    if (s & 0x80):
        y = -y

    return y


if __name__ == '__main__':
    if(len(sys.argv) < 4):
        print("bpatch: usage: python bpatch.py oldfile newfile patchfile")
    else:
        old_file_name = sys.argv[1]
        new_file_name = sys.argv[2]
        patch_file_name = sys.argv[3]
        patch(old_file_name, new_file_name, patch_file_name)



#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

from pymobiledevice.lockdown import LockdownClient
from pprint import pprint
import plistlib
import _thread
import time

# NP Client to device Notifications (post_notification)
NP_SYNC_WILL_START = "com.apple.itunes-mobdev.syncWillStart"
NP_SYNC_DID_START = "com.apple.itunes-mobdev.syncDidStart"
NP_SYNC_DID_FINISH = "com.apple.itunes-mobdev.syncDidFinish"
NP_SYNC_LOCK_REQUEST = "com.apple.itunes-mobdev.syncLockRequest"

# Device to NP Client Notifications (get_notification)
NP_SYNC_CANCEL_REQUEST = "com.apple.itunes-client.syncCancelRequest"
NP_SYNC_SUSPEND_REQUEST = "com.apple.itunes-client.syncSuspendRequest"
NP_SYNC_RESUME_REQUEST = "com.apple.itunes-client.syncResumeRequest"
NP_PHONE_NUMBER_CHANGED = "com.apple.mobile.lockdown.phone_number_changed"
NP_DEVICE_NAME_CHANGED = "com.apple.mobile.lockdown.device_name_changed"
NP_TIMEZONE_CHANGED = "com.apple.mobile.lockdown.timezone_changed"
NP_TRUSTED_HOST_ATTACHED = "com.apple.mobile.lockdown.trusted_host_attached"
NP_HOST_DETACHED = "com.apple.mobile.lockdown.host_detached"
NP_HOST_ATTACHED = "com.apple.mobile.lockdown.host_attached"
NP_REGISTRATION_FAILED = "com.apple.mobile.lockdown.registration_failed"
NP_ACTIVATION_STATE = "com.apple.mobile.lockdown.activation_state"
NP_BRICK_STATE = "com.apple.mobile.lockdown.brick_state"
NP_DISK_USAGE_CHANGED = "com.apple.mobile.lockdown.disk_usage_changed"
NP_DS_DOMAIN_CHANGED = "com.apple.mobile.data_sync.domain_changed"
NP_BACKUP_DOMAIN_CHANGED = "com.apple.mobile.backup.domain_changed"
NP_APP_INSTALLED = "com.apple.mobile.application_installed"
NP_APP_UNINSTALLED = "com.apple.mobile.application_uninstalled"
NP_DEV_IMAGE_MOUNTED = "com.apple.mobile.developer_image_mounted"
NP_ATTEMPTACTIVATION = "com.apple.springboard.attemptactivation"
NP_ITDBPREP_DID_END = "com.apple.itdbprep.notification.didEnd"
NP_LANGUAGE_CHANGED = "com.apple.language.changed"
NP_ADDRESS_BOOK_PREF_CHANGED = "com.apple.AddressBook.PreferenceChanged"


class NPClient(object):

    def __init__(self, lockdown=None, serviceName="com.apple.mobile.notification_proxy"):
        if lockdown:
            self.lockdown = lockdown
        else:
            self.lockdown = LockdownClient()
        self.service = self.lockdown.startService(serviceName)

    def stop_session(self):
        print("Disconecting...")
        self.service.close()

    def post_notification(self, notification):
        # Sends a notification to the device's notification_proxy.

        self.service.sendPlist({"Command": "PostNotification",
                                "Name": notification})

        self.service.sendPlist({"Command": "Shutdown"})
        res = self.service.recvPlist()
        pprint(res)
        if res:
            if res.get("Command") == "ProxyDeath":
                return res.get("Command")
            else:
                print("Got unknown NotificationProxy command %s" %
                      res.get("Command"))
                pprint(res)
        return None

    def observe_notification(self, notification):
        # Tells the device to send a notification on the specified event

        print("Observing %s" % notification)
        self.service.sendPlist({"Command": "ObserveNotification",
                                "Name": notification})

    def get_notification(self, notification):
        # Checks if a notification has been sent by the device

        res = self.service.recvPlist()
        if res:
            if res.get("Command") == "RelayNotification":
                if res.get("Name"):
                    return res.get("Name")

            elif res.get("Command") == "ProxyDeath":
                print("NotificationProxy died!")
            else:
                print("Got unknown NotificationProxy command %s" %
                      res.get("Command"))
                pprint(res)
        return None

    def notifier(self, name, args=None):

        if args == None:
            return None

        self.observe_notification(args.get("notification"))

        while args.get("running") == True:
            np_name = self.get_notification(args.get("notification"))
            if np_name:
                userdata = args.get("userdata")
                try:
                    _thread.start_new_thread(
                        args.get("callback"), (np_name, userdata, ))
                except:
                    print("Error: unable to start thread")

    def subscribe(self, notification, cb, data=None):

        np_data = {
            "running": True,
            "notification": notification,
            "callback": cb,
            "userdata": data,
        }

        try:
            _thread.start_new_thread(
                self.notifier, ("NotificationProxyNotifier_" + notification, np_data, ))
        except:
            print("Error: unable to start thread")

        while(1):
            time.sleep(1)


def cb_test(name, data=None):
    print("Got Notification >> %s" % name)
    print("Data:")
    pprint(data)


if __name__ == "__main__":
    np = NPClient()
    np.subscribe(NP_DEVICE_NAME_CHANGED, cb_test, data=None)



import base64


def chunks(l, n):
    return (l[i:i + n] for i in range(0, len(l), n))


def RSA_KEY_DER_to_PEM(data):
    a = ["-----BEGIN RSA PRIVATE KEY-----"]
    a.extend(chunks(base64.b64encode(data), 64))
    a.append("-----END RSA PRIVATE KEY-----")
    return "\n".join(a)



#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from pymobiledevice.lockdown import LockdownClient
from pprint import pprint
from pymobiledevice.afc import AFCClient, AFCShell
from optparse import OptionParser
import os
import warnings

warnings.warn(
    """The libraries upon which this program depends will soon be deprecated in
favor of the house_arrest.py and installation_proxy.py libraries.
See those files for example program written using the new libraries."""
)


def house_arrest(lockdown, applicationId):
    try:
        mis = lockdown.startService("com.apple.mobile.house_arrest")
    except:
        lockdown = LockdownClient()
        mis = lockdown.startService("com.apple.mobile.house_arrest")

    if mis == None:
        return
    mis.sendPlist({"Command": "VendDocuments", "Identifier": applicationId})
    res = mis.recvPlist()
    if res.get("Error"):
        print("Unable to Lookup the selected application: You probably trying to access to a system app...")
        return None
    return AFCClient(lockdown, service=mis)


def house_arrest_shell(lockdown, applicationId):
    afc = house_arrest(lockdown, applicationId)
    if afc:
        AFCShell(client=afc).cmdloop()

"""
"Install"
"Upgrade"
"Uninstall"
"Lookup"
"Browse"
"Archive"
"Restore"
"RemoveArchive"
"LookupArchives"
"CheckCapabilitiesMatch"

installd
if stat("/var/mobile/tdmtanf") => "TDMTANF Bypass" => SignerIdentity bypass
"""


def mobile_install(lockdown, ipaPath):
    # Start afc service & upload ipa
    afc = AFCClient(lockdown)
    afc.set_file_contents("/" + os.path.basename(ipaPath),
                          open(ipaPath, 'rb').read())
    mci = lockdown.startService("com.apple.mobile.installation_proxy")
    # print mci.sendPlist({"Command":"Archive","ApplicationIdentifier":
    # "com.joystickgenerals.STActionPig"})
    mci.sendPlist({"Command": "Install",
                   #"ApplicationIdentifier": "com.gotohack.JBChecker",
                   "PackagePath": os.path.basename(ipaPath)})
    while True:
        z = mci.recvPlist()
        if not z:
            break
        completion = z.get('PercentComplete')
        if completion:
            print('Installing, %s: %s %% Complete' %
                  (ipaPath, z['PercentComplete']))
        if z.get('Status') == 'Complete':
            print("Installation %s\n" % z['Status'])
            break


def list_apps(lockdown):
    mci = lockdown.startService("com.apple.mobile.installation_proxy")
    # print
    mci.sendPlist({"Command": "Lookup"})
    res = mci.recvPlist()
    for app in list(res["LookupResult"].values()):
        if app.get("ApplicationType") != "System":
            print(app["CFBundleIdentifier"], "=>", app.get("Container"))
        else:
            print(app["CFBundleIdentifier"], "=> N/A")


def get_apps_BundleID(lockdown, appType="User"):
    appList = []
    mci = lockdown.startService("com.apple.mobile.installation_proxy")
    mci.sendPlist({"Command": "Lookup"})
    res = mci.recvPlist()
    for app in list(res["LookupResult"].values()):
        if app.get("ApplicationType") == appType:
            appList.append(app["CFBundleIdentifier"])
        # else: #FIXME
        #    appList.append(app["CFBundleIdentifier"])
    mci.close()
    # pprint(appList)
    return appList


if __name__ == "__main__":
    parser = OptionParser(usage="%prog")
    parser.add_option("-l", "--list", dest="list", action="store_true", default=False,
                      help="List installed applications (non system apps)")
    parser.add_option("-a", "--app", dest="app", action="store", default=None,
                      metavar="APPID", help="Access application files with AFC")
    parser.add_option("-i", "--install", dest="installapp", action="store", default=None,
                      metavar="FILE", help="Install an application package")

    (options, args) = parser.parse_args()
    if options.list:
        lockdown = LockdownClient()
        list_apps(lockdown)
    elif options.app:
        lockdown = LockdownClient()
        house_arrest_shell(lockdown, options.app)
    elif options.installapp:
        lockdown = LockdownClient()
        mobile_install(lockdown, options.installapp)
    else:
        parser.print_help()



import glob
import plistlib
import os
import pickle
import gzip
from optparse import Option


def read_file(filename):
    f = open(filename, "rb")
    data = f.read()
    f.close()
    return data


def write_file(filename, data):
    f = open(filename, "wb")
    f.write(data)
    f.close()


def makedirs(dirs):
    try:
        os.makedirs(dirs)
    except:
        pass


def getHomePath(foldername, filename):
    home = os.path.expanduser('~')
    folderpath = os.path.join(home, foldername)
    if not os.path.exists(folderpath):
        makedirs(folderpath)
    return os.path.join(folderpath, filename)


def readHomeFile(foldername, filename):
    path = getHomePath(foldername, filename)
    if not os.path.exists(path):
        return None
    return read_file(path)

# return path to HOME+foldername+filename


def writeHomeFile(foldername, filename, data):
    filepath = getHomePath(foldername, filename)
    write_file(filepath, data)
    return filepath


def readPlist(filename):
    f = open(filename, "rb")
    f.read(16)
    f.close()
    return plistlib.readPlist(filename)


def parsePlist(s):
    return plistlib.readPlistFromString(s)

# http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size


def sizeof_fmt(num):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%d%s" % (num, x)
        num /= 1024.0

# http://www.5dollarwhitebox.org/drupal/node/84


def convert_bytes(bytes):
    bytes = float(bytes)
    if bytes >= 1099511627776:
        terabytes = bytes / 1099511627776
        size = '%.2fT' % terabytes
    elif bytes >= 1073741824:
        gigabytes = bytes / 1073741824
        size = '%.2fG' % gigabytes
    elif bytes >= 1048576:
        megabytes = bytes / 1048576
        size = '%.2fM' % megabytes
    elif bytes >= 1024:
        kilobytes = bytes / 1024
        size = '%.2fK' % kilobytes
    else:
        size = '%.2fb' % bytes
    return size


def xor_strings(a, b):
    r = ""
    for i in range(len(a)):
        r += chr(ord(a[i]) ^ ord(b[i]))
    return r

hex = lambda data: " ".join("%02X" % ord(i) for i in data)
ascii = lambda data: "".join(c if 31 < ord(c) < 127 else "." for c in data)


def hexdump(d):
    for i in range(0, len(d), 16):
        data = d[i:i + 16]
        print("%08X | %s | %s" % (i, hex(data).ljust(47), ascii(data)))


def search_plist(directory, matchDict):
    for p in map(os.path.normpath, glob.glob(directory + "/*.plist")):
        try:
            d = plistlib.readPlist(p)
            ok = True
            for k, v in list(matchDict.items()):
                if d.get(k) != v:
                    ok = False
                    break
            if ok:
                print("Using plist file %s" % p)
                return d
        except:
            continue


def save_pickle(filename, data):
    f = gzip.open(filename, "wb")
    pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)
    f.close()


def load_pickle(filename):
    f = gzip.open(filename, "rb")
    data = pickle.load(f)
    f.close()
    return data


class MultipleOption(Option):
    ACTIONS = Option.ACTIONS + ("extend",)
    STORE_ACTIONS = Option.STORE_ACTIONS + ("extend",)
    TYPED_ACTIONS = Option.TYPED_ACTIONS + ("extend",)
    ALWAYS_TYPED_ACTIONS = Option.ALWAYS_TYPED_ACTIONS + ("extend",)

    def take_action(self, action, dest, opt, value, values, parser):
        if action == "extend":
            values.ensure_value(dest, []).append(value)
        else:
            Option.take_action(self, action, dest, opt, value, values, parser)



#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from pymobiledevice.lockdown import LockdownClient
from optparse import OptionParser
from pprint import pprint
import plistlib
from pymobiledevice.util import read_file


class MobileConfigService(object):

    def __init__(self, lockdown):
        self.lockdown = lockdown
        self.service = lockdown.startService("com.apple.mobile.MCInstall")

    def GetProfileList(self):
        self.service.sendPlist({"RequestType": "GetProfileList"})
        res = self.service.recvPlist()
        if res.get("Status") != "Acknowledged":
            print("GetProfileList error")
            pprint(res)
            return
        return res

    def InstallProfile(self, s):
        #s = plistlib.writePlistToString(payload)
        self.service.sendPlist(
            {"RequestType": "InstallProfile", "Payload": plistlib.Data(s)})
        return self.service.recvPlist()

    def RemoveProfile(self, ident):
        profiles = self.GetProfileList()
        if not profiles:
            return
        if ident not in profiles["ProfileMetadata"]:
            print("Trying to remove not installed profile %s" % ident)
            return
        meta = profiles["ProfileMetadata"][ident]
        pprint(meta)
        data = plistlib.dumps({"PayloadType": "Configuration",
                               "PayloadIdentifier": ident,
                               "PayloadUUID": meta["PayloadUUID"],
                               "PayloadVersion": meta["PayloadVersion"]
                               })
        self.service.sendPlist(
            {"RequestType": "RemoveProfile", "ProfileIdentifier": plistlib.Data(data)})
        return self.service.recvPlist()


def main():
    parser = OptionParser(usage="%prog")
    parser.add_option("-l", "--list", dest="list", action="store_true",
                      default=False, help="List installed profiles")
    parser.add_option("-i", "--install", dest="install", action="store",
                      metavar="FILE", help="Install profile")
    parser.add_option("-r", "--remove", dest="remove", action="store",
                      metavar="IDENTIFIER", help="Remove profile")
    (options, args) = parser.parse_args()

    if not options.list and not options.install and not options.remove:
        parser.print_help()
        return
    lockdown = LockdownClient()
    mc = MobileConfigService(lockdown)

    if options.list:
        pprint(mc.GetProfileList())
    elif options.install:
        mc.InstallProfile(read_file(options.install))
    elif options.remove:
        pprint(mc.RemoveProfile(options.remove))

if __name__ == "__main__":
    main()




def print_table(title, headers, rows):
    widths = []

    for i in range(len(headers)):
        z = list(map(len, [row[i] for row in rows]))
        z.append(len(headers[i]))
        widths.append(max(z))

    width = sum(widths) + len(headers) + 1
    print("-" * width)
    print("|" + title.center(width - 2) + "|")
    print("-" * width)
    hline = "|"
    for i in range(len(headers)):
        hline += headers[i].ljust(widths[i]) + "|"
    print(hline)

    print("-" * width)
    for row in rows:
        line = "|"
        for i in range(len(row)):
            line += row[i].ljust(widths[i]) + "|"
        print(line)

    if len(rows) == 0:
        print("|" + "No entries".center(width - 2) + "|")
    print("-" * width)
    print("")



#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from pymobiledevice.lockdown import LockdownClient
from pprint import pprint
import plistlib
from time import gmtime, strftime
from optparse import OptionParser
import os
import plistlib


class screenshotr(object):

    def __init__(self, lockdown=None, serviceName='com.apple.mobile.screenshotr'):
        if lockdown:
            self.lockdown = lockdown
        else:
            self.lockdown = LockdownClient()
        # Starting Screenshot service
        self.service = self.lockdown.startService(serviceName)

        # hand check
        DLMessageVersionExchange = self.service.recvPlist()
        #assert len(DLMessageVersionExchange) == 2
        version_major = DLMessageVersionExchange[1]
        self.service.sendPlist(
            ["DLMessageVersionExchange", "DLVersionsOk", version_major])
        DLMessageDeviceReady = self.service.recvPlist()

    def stop_session(self):
        self.service.close()

    def take_screenshot(self):
        self.service.sendPlist(
            ['DLMessageProcessMessage', {'MessageType': 'ScreenShotRequest'}])
        res = self.service.recvPlist()

        assert len(res) == 2
        assert res[0] == "DLMessageProcessMessage"

        if res[1].get('MessageType') == 'ScreenShotReply':
            data = res[1]['ScreenShotData'].data
            return data
        return None

if __name__ == '__main__':
    parser = OptionParser(usage='%prog')
    parser.add_option('-p', '--path', dest='outDir', default=False,
                      help='Output Directory (default: . )', type='string')
    (options, args) = parser.parse_args()

    outPath = '.'
    if options.outDir:
        outPath = options.outDir

    screenshotr = screenshotr()
    data = screenshotr.take_screenshot()
    if data:
        filename = strftime('screenshot-%Y-%m-%d-%H-%M-%S.tif', gmtime())
        outPath = os.path.join(outPath, filename)
        print('Saving Screenshot at %s' % outPath)
        o = open(outPath, 'wb')
        o.write(data)



#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


import ssl
import struct
import plistlib
from re import sub
from ak_vendor.usbmux import default_mux as mux


class PlistService(object):

    def __init__(self, port, udid=None):
        self.port = port
        self.connect(udid)

    def connect(self, udid=None):
        mux.process(1.0)
        dev = None

        while not dev and mux.devices:
            mux.process(1.0)
            if udid:
                for d in mux.devices:
                    if d.serial == udid:
                        dev = d
                        print("Connecting to device: " + dev.serial)
            else:
                dev = mux.devices[0]
                print("Connecting to device: " + dev.serial)

        try:
            self.s = mux.connect(dev, self.port)
        except:
            raise Exception("Connexion to device port %d failed" % self.port)
        return dev.serial

    def close(self):
        self.s.close()

    def recv(self, length=4096):
        return self.s.recv(length)

    def send(self, data):
        try:
            self.s.send(data)
        except:
            print("Sending data to device failled")
            return -1
        return 0

    def sendRequest(self, data):
        res = None
        if self.sendPlist(data) >= 0:
            res = self.recvPlist()
        return res

    def recv_exact(self, l):
        data = b""
        while l > 0:
            d = self.recv(l)
            if not d or len(d) == 0:
                break
            data += d
            l -= len(d)
        return data

    def recv_raw(self):
        l = self.recv_exact(4)
        if not l or len(l) != 4:
            return
        l = struct.unpack(">L", l)[0]
        return self.recv_exact(l)

    def send_raw(self, data):
        return self.send(struct.pack(">L", len(data)) + data)

    def recvPlist(self):
        payload = self.recv_raw()
        if not payload:
            return
        if payload.startswith(b"bplist00"):
            return plistlib.loads(payload)
        elif payload.startswith(b"<?xml"):
            # HAX lockdown HardwarePlatform with null bytes
            payload = sub('[^\w<>\/ \-_0-9\"\'\\=\.\?\!\+]+',
                          '', payload.decode('utf-8')).encode('utf-8')
            return plistlib.loads(payload)
        else:
            raise Exception("recvPlist invalid data : %s" %
                            payload[:100].encode("hex"))

    def sendPlist(self, d):
        payload = plistlib.dumps(d)
        l = struct.pack(">L", len(payload))
        return self.send(l + payload)

    def ssl_start(self, keyfile, certfile):
        self.s = ssl.wrap_socket(
            self.s, keyfile, certfile, ssl_version=ssl.PROTOCOL_TLSv1)



#!/usr/bin/env python
# -*- coding: utf8 -*-
#
# $Id$
#
# Copyright (c) 2012-2014 "dark[-at-]gotohack.org"
#
# This file is part of pymobiledevice
#
# pymobiledevice is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import os
from pymobiledevice.lockdown import LockdownClient
from pprint import pprint
from pymobiledevice.afc import AFCClient, AFCShell
from optparse import OptionParser


class HouseArrestClient(AFCClient):

    def __init__(self, lockdown=None, serviceName="com.apple.mobile.house_arrest", service=None):
        self.serviceName = serviceName
        self.lockdown = lockdown if lockdown else LockdownClient()
        self.service = service if service else self.lockdown.startService(
            self.serviceName)

    def stop_session(self):
        print("Disconecting...")
        self.service.close()

    def send_command(self, applicationId, cmd="VendDocuments"):
        self.service.sendPlist({"Command": cmd, "Identifier": applicationId})
        res = self.service.recvPlist()

        if res.get("Error"):
            print(res["Error"])
            return None

    def shell(self, applicationId, cmd="VendDocuments"):
        res = self.send_command(applicationId, cmd="VendDocuments")
        if res:
            AFCShell(client=self.service).cmdloop()


if __name__ == "__main__":
    parser = OptionParser(usage="%prog -a  applicationId")
    parser.add_option("-a", "--application", dest="applicationId", default=False,
                      help="Application ID <com.apple.iBooks>", type="string")
    parser.add_option("-c", "--command", dest="cmd", default=False,
                      help="House_Arrest commands: ", type="string")

    (options, args) = parser.parse_args()
    h = HouseArrestClient()
    if options.cmd:
        h.shell(options.applicationId)
    else:
        h.shell(options.applicationId, cmd=options.cmd)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# vim: fenc=utf-8
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#

"""
File name: apis.py
Author: dhilipsiva <dhilipsiva@gmail.com>
Date created: 2016-06-19
"""

from os import path
# from pprint import pprint


def get_lockdown_and_service(udid):
    from pymobiledevice.lockdown import LockdownClient
    lockdown = LockdownClient(udid)
    service = lockdown.startService("com.apple.mobile.installation_proxy")
    return lockdown, service


def run_command(service, uuid, cmd):
    service.sendPlist(cmd)
    z = service.recvPlist()
    while 'PercentComplete' in z:
        if not z:
            break
        if z.get("Status") == "Complete":
            return z.get("Status")
        z = service.recvPlist()
    service.close()
    return z


def install_ipa(uuid, ipa_path):
    """
    docstring for install_ipa
    """
    from pymobiledevice.afc import AFCClient
    lockdown, service = get_lockdown_and_service(uuid)
    afc = AFCClient(lockdown=lockdown)
    afc.set_file_contents(
        path.basename(ipa_path), open(ipa_path, "rb").read())
    cmd = {"Command": "Install", "PackagePath": path.basename(ipa_path)}
    return run_command(service, uuid, cmd)


def uninstall_ipa(uuid, bundle_id):
    lockdown, service = get_lockdown_and_service(uuid)
    cmd = {"Command": "Uninstall", "ApplicationIdentifier": bundle_id}
    return run_command(service, uuid, cmd)


def list_ipas(uuid):
    lockdown, service = get_lockdown_and_service(uuid)
    cmd = {"Command": "Lookup"}
    result = run_command(service, uuid, cmd)
    apps_details = result.get("LookupResult")
    apps = []
    for app in apps_details:
        if apps_details[app]['ApplicationType'] == 'User':
            apps.append(app)
    return apps
